import time


class Timer(object):
    def __init__(self, log, description):
        self._log = log
        self._description = description

    def __enter__(self):
        self._start_time = time.clock()

    def __exit__(self, exc_type, exc_val, exc_tb):
        elapsed_time = (time.clock() - self._start_time) * 1.0e3
        self._log.info("{0}: Took {1} ms".format(
            self._description, elapsed_time))



"""Test reading of example TDMS files"""

import unittest
import sys
import logging
import binascii
import struct
import tempfile
from datetime import datetime

from nptdms import tdms


try:
    long
except NameError:
    # Python 3
    long = int


def string_hexlify(input_string):
    """Return hex string representation of string"""
    return binascii.hexlify(input_string.encode('utf-8')).decode('utf-8')


def hexlify_value(struct_type, value):
    """Return hex string representation of a value"""
    return binascii.hexlify(struct.pack(struct_type, value)).decode('utf-8')


def basic_segment():
    """Basic TDMS segment with one group and two channels"""

    toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList")
    metadata = (
        # Number of objects
        "03 00 00 00"
        # Length of the first object path
        "08 00 00 00"
        # Object path (/'Group')
        "2F 27 47 72"
        "6F 75 70 27"
        # Raw data index
        "FF FF FF FF"
        # Num properties
        "02 00 00 00"
        # Name length
        "04 00 00 00"
        # Property name (prop)
        "70 72 6F 70"
        # Property data type (string)
        "20 00 00 00"
        # Length of string value
        "05 00 00 00"
        # Value
        "76 61 6C 75 65"
        # Length of second property name
        "03 00 00 00"
        # Property name (num)
        "6E 75 6D"
        # Data type of property
        "03 00 00 00"
        # Value
        "0A 00 00 00"
        # Length of the second object path
        "13 00 00 00"
        # Second object path (/'Group'/'Channel1')
        "2F 27 47 72"
        "6F 75 70 27"
        "2F 27 43 68"
        "61 6E 6E 65"
        "6C 31 27"
        # Length of index information
        "14 00 00 00"
        # Raw data data type
        "03 00 00 00"
        # Dimension
        "01 00 00 00"
        # Number of raw datata values
        "02 00 00 00"
        "00 00 00 00"
        # Number of properties (0)
        "00 00 00 00"
        # Length of the third object path
        "13 00 00 00"
        # Third object path (/'Group'/'Channel2')
        "2F 27 47 72"
        "6F 75 70 27"
        "2F 27 43 68"
        "61 6E 6E 65"
        "6C 32 27"
        # Length of index information
        "14 00 00 00"
        # Raw data data type
        "03 00 00 00"
        # Dimension
        "01 00 00 00"
        # Number of data values
        "02 00 00 00"
        "00 00 00 00"
        # Set time properties for the second channel
        "02 00 00 00"
        "0F 00 00 00" +
        string_hexlify('wf_start_offset') +
        "0A 00 00 00" +
        hexlify_value("<d", 0.0) +
        "0C 00 00 00" +
        string_hexlify('wf_increment') +
        "0A 00 00 00" +
        hexlify_value("<d", 0.1))
    data = (
        # Data for segment
        "01 00 00 00"
        "02 00 00 00"
        "03 00 00 00"
        "04 00 00 00"
    )
    return (metadata, data, toc)


class TestFile(object):
    """Generate a TDMS file for testing"""

    def __init__(self):
        self._tempfile = tempfile.NamedTemporaryFile()
        self.file = self._tempfile.file
        self.data = bytes()

    def add_segment(self, metadata, data, toc=None, incomplete=False):
        metadata = self.to_bytes(metadata)
        data = self.to_bytes(data)
        if toc is not None:
            lead_in = b'TDSm'
            toc_mask = long(0)
            if "kTocMetaData" in toc:
                toc_mask = toc_mask | long(1) << 1
            if "kTocRawData" in toc:
                toc_mask = toc_mask | long(1) << 3
            if "kTocDAQmxRawData" in toc:
                toc_mask = toc_mask | long(1) << 7
            if "kTocInterleavedData" in toc:
                toc_mask = toc_mask | long(1) << 5
            if "kTocBigEndian" in toc:
                toc_mask = toc_mask | long(1) << 6
            if "kTocNewObjList" in toc:
                toc_mask = toc_mask | long(1) << 2
            lead_in += struct.pack('<i', toc_mask)
            lead_in += self.to_bytes("69 12 00 00")
            next_segment_offset = len(metadata) + len(data)
            raw_data_offset = len(metadata)
            if incomplete:
                lead_in += self.to_bytes('FF' * 8)
            else:
                lead_in += struct.pack('<Q', next_segment_offset)
            lead_in += struct.pack('<Q', raw_data_offset)
        else:
            lead_in = b''
        self.data += lead_in + metadata + data

    def to_bytes(self, hex_data):
        return binascii.unhexlify(
            hex_data.replace(" ", "").replace("\n", "").encode('utf-8'))

    def load(self, *args, **kwargs):
        self.file.write(self.data)
        self.file.seek(0)
        return tdms.TdmsFile(self.file, *args, **kwargs)


class TDMSTestClass(unittest.TestCase):
    def setUp(self):
        logging.getLogger(tdms.__name__).setLevel(logging.DEBUG)

    def test_data_read(self):
        """Test reading data"""

        test_file = TestFile()
        test_file.add_segment(*basic_segment())
        tdmsData = test_file.load()

        data = tdmsData.channel_data("Group", "Channel1")
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0], 1)
        self.assertEqual(data[1], 2)
        data = tdmsData.channel_data("Group", "Channel2")
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0], 3)
        self.assertEqual(data[1], 4)

    def test_property_read(self):
        """Test reading an object property"""

        test_file = TestFile()
        test_file.add_segment(*basic_segment())
        tdmsData = test_file.load()

        object = tdmsData.object("Group")
        self.assertEqual(object.property("num"), 10)

    def test_no_metadata_segment(self):
        """Add a segment with two channels, then a second
        segment with the same metadata as before,
        so there is only the lead in and binary data"""

        test_file = TestFile()
        (metadata, data, toc) = basic_segment()
        test_file.add_segment(metadata, data, toc)
        data = (
            "05 00 00 00"
            "06 00 00 00"
            "07 00 00 00"
            "08 00 00 00"
        )
        toc = ("kTocRawData")
        test_file.add_segment('', data, toc)
        tdmsData = test_file.load()

        data = tdmsData.channel_data("Group", "Channel1")
        self.assertEqual(len(data), 4)
        self.assertTrue(all(data == [1, 2, 5, 6]))
        data = tdmsData.channel_data("Group", "Channel2")
        self.assertEqual(len(data), 4)
        self.assertTrue(all(data == [3, 4, 7, 8]))

    def test_no_metadata_object(self):
        """Re-use an object without setting any new metadata and
        re-using the data structure"""

        test_file = TestFile()
        (metadata, data, toc) = basic_segment()
        test_file.add_segment(metadata, data, toc)
        data = (
            "05 00 00 00"
            "06 00 00 00"
            "07 00 00 00"
            "08 00 00 00"
        )
        toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList")
        # Use same object list, but set raw data index to 0
        metadata = (
            # Number of objects
            "03 00 00 00"
            # Length of the first object path
            "08 00 00 00"
            # Object path (/'Group')
            "2F 27 47 72"
            "6F 75 70 27"
            # Raw data index
            "FF FF FF FF"
            # Num properties
            "00 00 00 00"
            # Length of the second object path
            "13 00 00 00"
            # Second object path (/'Group'/'Channel1')
            "2F 27 47 72"
            "6F 75 70 27"
            "2F 27 43 68"
            "61 6E 6E 65"
            "6C 31 27"
            # Raw data index meaning repeat previous data structure
            "00 00 00 00"
            # Number of properties (0)
            "00 00 00 00"
            # Length of the third object path
            "13 00 00 00"
            # Third object path (/'Group'/'Channel2')
            "2F 27 47 72"
            "6F 75 70 27"
            "2F 27 43 68"
            "61 6E 6E 65"
            "6C 32 27"
            # Raw data index meaning repeat previous data structure
            "00 00 00 00"
            # Number of properties
            "00 00 00 00")
        test_file.add_segment(metadata, data, toc)
        tdmsData = test_file.load()

        data = tdmsData.channel_data("Group", "Channel1")
        self.assertEqual(len(data), 4)
        self.assertTrue(all(data == [1, 2, 5, 6]))
        data = tdmsData.channel_data("Group", "Channel2")
        self.assertEqual(len(data), 4)
        self.assertTrue(all(data == [3, 4, 7, 8]))

    def test_new_channel(self):
        """Add a new voltage channel, with the other two channels
        remaining unchanged, so only the new channel is in metadata section"""

        test_file = TestFile()
        (metadata, data, toc) = basic_segment()
        test_file.add_segment(metadata, data, toc)
        toc = ("kTocMetaData", "kTocRawData")
        metadata = (
            # Number of objects
            "01 00 00 00"
            # Length of the third object path
            "12 00 00 00")
        metadata += string_hexlify("/'Group'/'Voltage'")
        metadata += (
            # Length of index information
            "14 00 00 00"
            # Raw data data type
            "03 00 00 00"
            # Dimension
            "01 00 00 00"
            # Number of raw datata values
            "02 00 00 00"
            "00 00 00 00"
            # Number of properties (0)
            "00 00 00 00")
        data = (
            "05 00 00 00"
            "06 00 00 00"
            "07 00 00 00"
            "08 00 00 00"
            "09 00 00 00"
            "0A 00 00 00"
        )
        test_file.add_segment(metadata, data, toc)
        tdmsData = test_file.load()

        data = tdmsData.channel_data("Group", "Channel1")
        self.assertEqual(len(data), 4)
        self.assertTrue(all(data == [1, 2, 5, 6]))
        data = tdmsData.channel_data("Group", "Channel2")
        self.assertEqual(len(data), 4)
        self.assertTrue(all(data == [3, 4, 7, 8]))
        data = tdmsData.channel_data("Group", "Voltage")
        self.assertEqual(len(data), 2)
        self.assertTrue(all(data == [9, 10]))

    def test_larger_channel(self):
        """In the second segment, increase the channel size
        of one channel"""

        test_file = TestFile()
        (metadata, data, toc) = basic_segment()
        test_file.add_segment(metadata, data, toc)
        toc = ("kTocMetaData", "kTocRawData")
        metadata = (
            # Number of objects
            "01 00 00 00"
            # Length of the object path
            "13 00 00 00")
        metadata += string_hexlify("/'Group'/'Channel2'")
        metadata += (
            # Length of index information
            "14 00 00 00"
            # Raw data data type
            "03 00 00 00"
            # Dimension
            "01 00 00 00"
            # Number of raw datata values
            "04 00 00 00"
            "00 00 00 00"
            # Number of properties (0)
            "00 00 00 00")
        data = (
            "05 00 00 00"
            "06 00 00 00"
            "07 00 00 00"
            "08 00 00 00"
            "09 00 00 00"
            "0A 00 00 00"
        )
        test_file.add_segment(metadata, data, toc)
        tdmsData = test_file.load()
        data = tdmsData.channel_data("Group", "Channel1")
        self.assertEqual(len(data), 4)
        self.assertTrue(all(data == [1, 2, 5, 6]))
        data = tdmsData.channel_data("Group", "Channel2")
        self.assertEqual(len(data), 6)
        self.assertTrue(all(data == [3, 4, 7, 8, 9, 10]))

    def test_remove_channel(self):
        """In the second segment, remove a channel.
        We need to write a new object list in this case"""

        test_file = TestFile()
        (metadata, data, toc) = basic_segment()
        test_file.add_segment(metadata, data, toc)
        # Keep toc as it was before, with new object list set
        metadata = (
            # Number of objects
            "01 00 00 00"
            # Length of the object path
            "13 00 00 00")
        metadata += string_hexlify("/'Group'/'Channel1'")
        metadata += (
            # Length of index information
            "14 00 00 00"
            # Raw data data type
            "03 00 00 00"
            # Dimension
            "01 00 00 00"
            # Number of raw datata values
            "02 00 00 00"
            "00 00 00 00"
            # Number of properties (0)
            "00 00 00 00")
        data = (
            "05 00 00 00"
            "06 00 00 00"
        )
        test_file.add_segment(metadata, data, toc)
        tdmsData = test_file.load()
        data = tdmsData.channel_data("Group", "Channel1")
        self.assertEqual(len(data), 4)
        self.assertTrue(all(data == [1, 2, 5, 6]))
        data = tdmsData.channel_data("Group", "Channel2")
        self.assertEqual(len(data), 2)
        self.assertTrue(all(data == [3, 4]))

    def test_no_lead_in(self):
        """Add segment and then a repeated segment without
        any lead in or metadata, so data is read in chunks"""

        test_file = TestFile()
        (metadata, data, toc) = basic_segment()
        data = data + (
            "05 00 00 00"
            "06 00 00 00"
            "07 00 00 00"
            "08 00 00 00"
        )
        test_file.add_segment(metadata, data, toc)
        tdmsData = test_file.load()

        data = tdmsData.channel_data("Group", "Channel1")
        self.assertEqual(len(data), 4)
        self.assertTrue(all(data == [1, 2, 5, 6]))
        data = tdmsData.channel_data("Group", "Channel2")
        self.assertEqual(len(data), 4)
        self.assertTrue(all(data == [3, 4, 7, 8]))

    def test_interleaved(self):
        """Test reading interleaved data"""

        test_file = TestFile()
        (metadata, data, toc) = basic_segment()
        toc = toc + ("kTocInterleavedData", )
        test_file.add_segment(metadata, data, toc)
        tdmsData = test_file.load()

        data = tdmsData.channel_data("Group", "Channel1")
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0], 1)
        self.assertEqual(data[1], 3)
        data = tdmsData.channel_data("Group", "Channel2")
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0], 2)
        self.assertEqual(data[1], 4)

    def test_timestamp_data(self):
        """Test reading contiguous and interleaved timestamp data,
        which isn't read by numpy"""

        times = [
            datetime(2012, 8, 23, 0, 0, 0, 123, tzinfo=tdms.timezone),
            datetime(2012, 8, 23, 1, 2, 3, 456, tzinfo=tdms.timezone),
            datetime(2012, 8, 23, 12, 0, 0, 0, tzinfo=tdms.timezone),
            datetime(2012, 8, 23, 12, 2, 3, 9999, tzinfo=tdms.timezone),
        ]

        def total_seconds(td):
            # timedelta.total_seconds() only added in 2.7
            return td.seconds + td.days * 24 * 3600

        seconds = [
            total_seconds(
                t - datetime(1904, 1, 1, 0, 0, 0, tzinfo=tdms.timezone))
            for t in times]
        fractions = [
            int(float(t.microsecond) * 2 ** 58 / 5 ** 6)
            for t in times]

        metadata = (
            # Number of objects
            "02 00 00 00"
            # Length of the object path
            "17 00 00 00")
        metadata += string_hexlify("/'Group'/'TimeChannel1'")
        metadata += (
            # Length of index information
            "14 00 00 00"
            # Raw data data type
            "44 00 00 00"
            # Dimension
            "01 00 00 00"
            # Number of raw datata values
            "02 00 00 00"
            "00 00 00 00"
            # Number of properties (0)
            "00 00 00 00")
        metadata += (
            "17 00 00 00")
        metadata += string_hexlify("/'Group'/'TimeChannel2'")
        metadata += (
            # Length of index information
            "14 00 00 00"
            # Raw data data type
            "44 00 00 00"
            # Dimension
            "01 00 00 00"
            # Number of raw datata values
            "02 00 00 00"
            "00 00 00 00"
            # Number of properties (0)
            "00 00 00 00")
        data = ""
        for f, s in zip(fractions, seconds):
            data += hexlify_value("<Q", f)
            data += hexlify_value("<q", s)

        def assertTimeEqual(a, b):
            """Check times are equal, allowing for small difference
            in microseconds due to floating point math being used
            """

            self.assertEqual(
                a.replace(microsecond=0), b.replace(microsecond=0))
            assert(abs(a.microsecond - b.microsecond) < 10)

        test_file = TestFile()
        toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList")
        test_file.add_segment(metadata, data, toc)
        tdmsData = test_file.load()
        channel_data = tdmsData.channel_data("Group", "TimeChannel1")
        self.assertEqual(len(channel_data), 2)
        assertTimeEqual(channel_data[0], times[0])
        assertTimeEqual(channel_data[1], times[1])
        # Read fraction of second
        channel_data = tdmsData.channel_data("Group", "TimeChannel2")
        self.assertEqual(len(channel_data), 2)
        assertTimeEqual(channel_data[0], times[2])
        assertTimeEqual(channel_data[1], times[3])

        # Now test it interleaved
        toc = toc + ("kTocInterleavedData", )
        test_file = TestFile()
        test_file.add_segment(metadata, data, toc)
        tdmsData = test_file.load()
        channel_data = tdmsData.channel_data("Group", "TimeChannel1")
        self.assertEqual(len(channel_data), 2)
        assertTimeEqual(channel_data[0], times[0])
        assertTimeEqual(channel_data[1], times[2])
        channel_data = tdmsData.channel_data("Group", "TimeChannel2")
        self.assertEqual(len(channel_data), 2)
        assertTimeEqual(channel_data[0], times[1])
        assertTimeEqual(channel_data[1], times[3])

    def test_time_track(self):
        """Add a time track to waveform data"""

        test_file = TestFile()
        (metadata, data, toc) = basic_segment()
        test_file.add_segment(metadata, data, toc)
        tdmsData = test_file.load()

        obj = tdmsData.object("Group", "Channel2")
        time = obj.time_track()
        self.assertEqual(len(time), len(obj.data))
        epsilon = 1.0E-15
        self.assertTrue(abs(time[0]) < epsilon)
        self.assertTrue(abs(time[1] - 0.1) < epsilon)

    def test_no_data_section(self):
        """kTocRawData is set but data length is zero

        Keep first segment the same but add a second
        segment with no data."""

        test_file = TestFile()
        (metadata, data, toc) = basic_segment()
        test_file.add_segment(metadata, data, toc)
        toc = ("kTocMetaData", "kTocRawData")
        metadata = (
            # Number of objects
            "02 00 00 00"
            # Length of the object path
            "13 00 00 00")
        metadata += string_hexlify("/'Group'/'Channel1'")
        metadata += (
            # Length of index information
            "14 00 00 00"
            # Raw data data type
            "03 00 00 00"
            # Dimension
            "01 00 00 00"
            # Number of raw datata values
            "00 00 00 00"
            "00 00 00 00"
            # Number of properties (0)
            "00 00 00 00")
        metadata += (
            # Length of the object path
            "13 00 00 00")
        metadata += string_hexlify("/'Group'/'Channel2'")
        metadata += (
            # Length of index information
            "14 00 00 00"
            # Raw data data type
            "03 00 00 00"
            # Dimension
            "01 00 00 00"
            # Number of raw datata values
            "00 00 00 00"
            "00 00 00 00"
            # Number of properties (0)
            "00 00 00 00")
        data = ""
        test_file.add_segment(metadata, data, toc)
        tdmsData = test_file.load()
        data = tdmsData.channel_data("Group", "Channel1")
        self.assertEqual(len(data), 2)
        self.assertTrue(all(data == [1, 2]))
        data = tdmsData.channel_data("Group", "Channel2")
        self.assertEqual(len(data), 2)
        self.assertTrue(all(data == [3, 4]))

    def test_repeated_object_without_data(self):
        """Repeated objects with no data in new segment

        A new object is also added with new data in order
        to trigger a bug with the chunk size calculation.
        """

        test_file = TestFile()
        (metadata, data, toc) = basic_segment()
        test_file.add_segment(metadata, data, toc)
        toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList")
        metadata = (
            # Number of objects
            "03 00 00 00"
            # Length of the object path
            "13 00 00 00")
        metadata += string_hexlify("/'Group'/'Channel1'")
        metadata += (
            # Raw data index
            "FF FF FF FF"
            # Number of properties (0)
            "00 00 00 00")
        metadata += (
            # Length of the object path
            "13 00 00 00")
        metadata += string_hexlify("/'Group'/'Channel2'")
        metadata += (
            # Raw data index
            "FF FF FF FF"
            # Number of properties (0)
            "00 00 00 00"
            # Length of the third object path
            "13 00 00 00")
        metadata += string_hexlify("/'Group'/'Channel3'")
        metadata += (
            # Length of index information
            "14 00 00 00"
            # Raw data data type
            "03 00 00 00"
            # Dimension
            "01 00 00 00"
            # Number of raw datata values
            "02 00 00 00"
            "00 00 00 00"
            # Number of properties (0)
            "00 00 00 00")
        data = (
            "01 00 00 00"
            "02 00 00 00"
        )
        test_file.add_segment(metadata, data, toc)
        tdmsData = test_file.load()
        data = tdmsData.channel_data("Group", "Channel1")
        self.assertEqual(len(data), 2)
        self.assertTrue(all(data == [1, 2]))
        data = tdmsData.channel_data("Group", "Channel2")
        self.assertEqual(len(data), 2)
        self.assertTrue(all(data == [3, 4]))
        data = tdmsData.channel_data("Group", "Channel3")
        self.assertEqual(len(data), 2)
        self.assertTrue(all(data == [1, 2]))

    def test_memmapped_read(self):
        """Test reading data into memmapped arrays"""

        test_file = TestFile()
        test_file.add_segment(*basic_segment())
        tdmsData = test_file.load(memmap_dir=tempfile.gettempdir())

        data = tdmsData.channel_data("Group", "Channel1")
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0], 1)
        self.assertEqual(data[1], 2)
        data = tdmsData.channel_data("Group", "Channel2")
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0], 3)
        self.assertEqual(data[1], 4)

    def test_incomplete_data(self):
        """Test incomplete last segment, eg. if LabView crashed"""

        test_file = TestFile()
        (metadata, data, toc) = basic_segment()
        test_file.add_segment(metadata, data, toc)
        # Add second, incomplete segment
        test_file.add_segment(metadata, data, toc, incomplete=True)
        tdmsData = test_file.load()

        # Eventually we might want to attempt to read the data
        # from the incomplete segment, but for now just make
        # sure we can read the data from previous segments
        data = tdmsData.channel_data("Group", "Channel1")
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0], 1)
        self.assertEqual(data[1], 2)
        data = tdmsData.channel_data("Group", "Channel2")
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0], 3)
        self.assertEqual(data[1], 4)

    def test_string_data(self):
        """Test reading a file with string data"""

        strings = ["abcdefg", "qwertyuiop"]

        test_file = TestFile()
        toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList")
        metadata = (
            # Number of objects
            "01 00 00 00"
            # Length of the object path
            "18 00 00 00")
        metadata += string_hexlify("/'Group'/'StringChannel'")
        metadata += (
            # Length of index information
            "1C 00 00 00"
            # Raw data data type
            "20 00 00 00"
            # Dimension
            "01 00 00 00"
            # Number of raw datata values
            "02 00 00 00"
            "00 00 00 00"
            # Number of bytes in data
            "19 00 00 00"
            "00 00 00 00"
            # Number of properties (0)
            "00 00 00 00")
        data = (
            "07 00 00 00"  # index to after first string
            "11 00 00 00"  # index to after second string
        )
        for string in strings:
            data += string_hexlify(string)
        test_file.add_segment(metadata, data, toc)
        tdmsData = test_file.load()

        data = tdmsData.channel_data("Group", "StringChannel")
        self.assertEqual(len(data), len(strings))
        for expected, read in zip(strings, data):
            self.assertEqual(expected, read)

    def test_slash_and_space_in_name(self):
        """Test name like '01/02/03 something'"""

        group_1_name = "01/02/03 something"
        channel_1_name = "04/05/06 another thing"
        group_2_name = "01/02/03 a"
        channel_2_name = "04/05/06 b"

        test_file = TestFile()

        toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList")

        # Number of objects
        metadata = "04 00 00 00"

        for group in [group_1_name, group_2_name]:
            path = "/'{0}'".format(group)
            metadata += hexlify_value('<l', len(path))
            metadata += string_hexlify(path)
            metadata += (
                # Raw data index
                "FF FF FF FF"
                # Number of properties (0)
                "00 00 00 00"
            )
        for (group, channel) in [
                (group_1_name, channel_1_name),
                (group_2_name, channel_2_name)]:
            path = "/'{0}'/'{1}'".format(group, channel)
            metadata += hexlify_value('<l', len(path))
            metadata += string_hexlify(path)
            metadata += (
                # Length of index information
                "14 00 00 00"
                # Raw data data type
                "03 00 00 00"
                # Dimension
                "01 00 00 00"
                # Number of raw datata values
                "02 00 00 00"
                "00 00 00 00"
                # Number of properties (0)
                "00 00 00 00"
            )

        data = (
            # Data for segment
            "01 00 00 00"
            "02 00 00 00"
            "03 00 00 00"
            "04 00 00 00"
        )

        test_file.add_segment(metadata, data, toc)
        tdmsData = test_file.load()

        self.assertEqual(len(tdmsData.groups()), 2)
        self.assertEqual(len(tdmsData.group_channels(group_1_name)), 1)
        self.assertEqual(len(tdmsData.group_channels(group_2_name)), 1)
        data_1 = tdmsData.channel_data(group_1_name, channel_1_name)
        self.assertEqual(len(data_1), 2)
        data_2 = tdmsData.channel_data(group_2_name, channel_2_name)
        self.assertEqual(len(data_2), 2)


if __name__ == '__main__':
    unittest.main()



from __future__ import print_function

from argparse import ArgumentParser
import logging

from nptdms import tdms


def main():
    parser = ArgumentParser(
        description="List the contents of a LabView TDMS file.")
    parser.add_argument(
        '-p', '--properties', action="store_true",
        help="Include channel properties.")
    parser.add_argument(
        '-d', '--debug', action="store_true",
        help="Print debugging information to stderr.")
    parser.add_argument(
        'tdms_file',
        help="TDMS file to read.")
    args = parser.parse_args()

    if args.debug:
        logging.getLogger(tdms.__name__).setLevel(logging.DEBUG)

    tdmsfile = tdms.TdmsFile(args.tdms_file)

    level = 0
    root = tdmsfile.object()
    display('/', level)
    if args.properties:
        display_properties(root, level)
    for group in tdmsfile.groups():
        level = 1
        try:
            group_obj = tdmsfile.object(group)
            display("%s" % group_obj.path, level)
            if args.properties:
                display_properties(group_obj, level)
        except KeyError:
            # It is possible to have a group without an object
            display("/'%s'" % group, level)
        for channel in tdmsfile.group_channels(group):
            level = 2
            display("%s" % channel.path, level)
            if args.properties:
                level = 3
                display("data type: %s" % channel.data_type.name, level)
                display_properties(channel, level)


def display_properties(tdms_object, level):
    if tdms_object.properties:
        display("properties:", level)
        for prop, val in tdms_object.properties.items():
            display("%s: %s" % (prop, val), level)


def display(s, level):
    print("%s%s" % (" " * 2 * level, s))



"""Python module for reading TDMS files produced by LabView"""

import itertools
import logging
import struct
import sys
from collections import namedtuple
try:
    from collections import OrderedDict
except ImportError:
    try:
        # ordereddict available on pypi for Python < 2.7
        from ordereddict import OrderedDict
    except ImportError:
        # Otherwise fall back on normal dict
        OrderedDict = dict
from copy import copy
import numpy as np
from datetime import datetime, timedelta
import tempfile
try:
    import pytz
except ImportError:
    pytz = None

from nptdms.utils import Timer


log = logging.getLogger(__name__)
logging.basicConfig()
log.setLevel(logging.WARNING)
# To adjust the log level for this module from a script, use eg:
# logging.getLogger(tdms.__name__).setLevel(logging.DEBUG)

try:
    long
    zip_longest = itertools.izip_longest
except NameError:
    # Python 3
    long = int
    zip_longest = itertools.zip_longest
tocProperties = {
    'kTocMetaData': (long(1) << 1),
    'kTocRawData': (long(1) << 3),
    'kTocDAQmxRawData': (long(1) << 7),
    'kTocInterleavedData': (long(1) << 5),
    'kTocBigEndian': (long(1) << 6),
    'kTocNewObjList': (long(1) << 2)
}

# Class for describing data types, with data type name,
# identifier used by struct module, the size in bytes to read and the
# numpy data type where applicable/implemented
DataType = namedtuple(
    "DataType", ('name', 'struct', 'length', 'nptype'))

tdsDataTypes = dict(enumerate((
    DataType('tdsTypeVoid', None, 0, None),
    DataType('tdsTypeI8', 'b', 1, np.int8),
    DataType('tdsTypeI16', 'h', 2, np.int16),
    DataType('tdsTypeI32', 'l', 4, np.int32),
    DataType('tdsTypeI64', 'q', 8, np.int64),
    DataType('tdsTypeU8', 'B', 1, np.uint8),
    DataType('tdsTypeU16', 'H', 2, np.uint16),
    DataType('tdsTypeU32', 'L', 4, np.uint32),
    DataType('tdsTypeU64', 'Q', 8, np.uint64),
    DataType('tdsTypeSingleFloat', 'f', 4, np.single),
    DataType('tdsTypeDoubleFloat', 'd', 8, np.double),
    DataType('tdsTypeExtendedFloat', None, None, None),
    DataType('tdsTypeDoubleFloatWithUnit', None, 8, None),
    DataType('tdsTypeExtendedFloatWithUnit', None, None, None)
)))

tdsDataTypes.update({
    0x19: DataType('tdsTypeSingleFloatWithUnit', None, 4, None),
    0x20: DataType('tdsTypeString', None, None, None),
    0x21: DataType('tdsTypeBoolean', 'b', 1, np.bool8),
    0x44: DataType('tdsTypeTimeStamp', 'Qq', 16, None),
    0xFFFFFFFF: DataType('tdsTypeDAQmxRawData', None, None, None)
})


if pytz:
    # Use UTC time zone if pytz is installed
    timezone = pytz.utc
else:
    timezone = None


def read_string(file):
    """Read a string from a tdms file

    For reading strings in the meta data section that start with
    the string length, will not work in the data section"""

    s = file.read(4)
    length = struct.unpack("<L", s)[0]
    return file.read(length).decode('utf-8')


def read_type(file, data_type, endianness):
    """Read a value from the file of the specified data type"""

    if data_type.name == 'tdsTypeTimeStamp':
        # Time stamps are stored as number of seconds since
        # 01/01/1904 00:00:00.00 UTC, ignoring leap seconds,
        # and number of 2^-64 fractions of a second.
        # Note that the TDMS epoch is not the Unix epoch.
        s = file.read(data_type.length)
        (s_frac, s) = struct.unpack('%s%s' % (endianness, data_type.struct), s)
        tdms_start = datetime(1904, 1, 1, 0, 0, 0, tzinfo=timezone)
        ms = float(s_frac) * 5 ** 6 / 2 ** 58
        # Adding timedelta with seconds ignores leap
        # seconds, so this is correct
        return tdms_start + timedelta(seconds=s) + timedelta(microseconds=ms)
    elif None not in (data_type.struct, data_type.length):
        s = file.read(data_type.length)
        return struct.unpack('%s%s' % (endianness, data_type.struct), s)[0]
    else:
        raise ValueError("Unsupported data type to read, %s." % data_type.name)


class TdmsFile(object):
    """Reads and stores data from a TDMS file.

    :ivar objects: A dictionary of objects in the TDMS file, where the keys are
        the object paths.

    """

    def __init__(self, file, memmap_dir=None):
        """Initialise a new TDMS file object, reading all data.

        :param file: Either the path to the tdms file to read or an already
            opened file.
        :param memmap_dir: The directory to store memmapped data files in,
            or None to read data into memory. The data files are created
            as temporary files and are deleted when the channel data is no
            longer used. tempfile.gettempdir() can be used to get the default
            temporary file directory.
        """

        self.segments = []
        self.objects = OrderedDict()
        self.memmap_dir = memmap_dir

        if hasattr(file, "read"):
            # Is a file
            self._read_segments(file)
        else:
            # Is path to a file
            with open(file, 'rb') as tdms_file:
                self._read_segments(tdms_file)

    def _read_segments(self, tdms_file):
        with Timer(log, "Read metadata"):
            # Read metadata first to work out how much space we need
            previous_segment = None
            while True:
                try:
                    segment = _TdmsSegment(tdms_file)
                except EOFError:
                    # We've finished reading the file
                    break
                segment.read_metadata(
                    tdms_file, self.objects, previous_segment)

                self.segments.append(segment)
                previous_segment = segment
                if segment.next_segment_pos is None:
                    break
                else:
                    tdms_file.seek(segment.next_segment_pos)

        with Timer(log, "Allocate space"):
            # Allocate space for data
            for object in self.objects.values():
                object._initialise_data(memmap_dir=self.memmap_dir)

        with Timer(log, "Read data"):
            # Now actually read all the data
            for segment in self.segments:
                segment.read_raw_data(tdms_file)

    def _path(self, *args):
        """Convert group and channel to object path"""

        return ('/' + '/'.join(
                ["'" + arg.replace("'", "''") + "'" for arg in args]))

    def _path_components(self, path):
        """Convert a path into group and channel name components"""

        def yield_components(path):
            # Iterate over each character and the next character
            chars = zip_longest(path, path[1:])
            try:
                # Iterate over components
                while True:
                    c, n = next(chars)
                    if c != '/':
                        raise ValueError("Invalid path, expected \"/\"")
                    elif (n is not None and n != "'"):
                        raise ValueError("Invalid path, expected \"'\"")
                    else:
                        # Consume "'" or raise StopIteration if at the end
                        next(chars)
                    component = []
                    # Iterate over characters in component name
                    while True:
                        c, n = next(chars)
                        if c == "'" and n == "'":
                            component += "'"
                            # Consume second "'"
                            next(chars)
                        elif c == "'":
                            yield "".join(component)
                            break
                        else:
                            component += c
            except StopIteration:
                return

        return list(yield_components(path))

    def object(self, *path):
        """Get a TDMS object from the file

        :param path: The object group and channel. Providing no channel
            returns a group object, and providing no channel or group
            will return the root object.
        :rtype: :class:`TdmsObject`

        For example, to get the root object::

            object()

        To get a group::

            object("group_name")

        To get a channel::

            object("group_name", "channel_name")
        """

        object_path = self._path(*path)
        try:
            return self.objects[object_path]
        except KeyError:
            raise KeyError("Invalid object path: %s" % object_path)

    def groups(self):
        """Return the names of groups in the file

        Note that there is not necessarily a TDMS object associated with
        each group name.

        :rtype: List of strings.

        """

        # Split paths into components and take the first (group) component.
        object_paths = (
            self._path_components(path)
            for path in self.objects)
        group_names = (path[0] for path in object_paths if len(path) > 0)

        # Use an ordered dict as an ordered set to find unique
        # groups in order.
        groups_set = OrderedDict()
        for group in group_names:
            groups_set[group] = None
        return list(groups_set)

    def group_channels(self, group):
        """Returns a list of channel objects for the given group

        :param group: Name of the group to get channels for.
        :rtype: List of :class:`TdmsObject` objects.

        """

        path = self._path(group)
        return [
            self.objects[p]
            for p in self.objects
            if p.startswith(path + '/')]

    def channel_data(self, group, channel):
        """Get the data for a channel

        :param group: The name of the group the channel is in.
        :param channel: The name of the channel to get data for.
        :returns: The channel data.
        :rtype: NumPy array.

        """

        return self.object(group, channel).data

    def as_dataframe(self, time_index=False, absolute_time=False):
        """
        Converts the TDMS file to a DataFrame

        :param time_index: Whether to include a time index for the dataframe.
        :param absolute_time: If time_index is true, whether the time index
            values are absolute times or relative to the start time.
        :return: The full TDMS file data.
        :rtype: pandas.DataFrame
        """

        import pandas as pd

        temp = {}
        for key, value in self.objects.items():
            if value.has_data:
                index = value.time_track(absolute_time) if time_index else None
                temp[key] = pd.Series(data=value.data, index=index)
        return pd.DataFrame.from_dict(temp)


class _TdmsSegment(object):

    __slots__ = [
        'position', 'num_chunks', 'ordered_objects', 'toc', 'version',
        'next_segment_offset', 'next_segment_pos',
        'raw_data_offset', 'data_position']

    def __init__(self, f):
        """Read the lead in section of a segment"""

        self.position = f.tell()
        self.num_chunks = 0
        # A list of _TdmsSegmentObject
        self.ordered_objects = []

        # First four bytes should be TDSm
        try:
            s = f.read(4).decode('utf-8')
        except UnicodeDecodeError:
            raise ValueError("Segment does not start with TDSm")
        if s == '':
            raise EOFError
        if s != 'TDSm':
            raise ValueError(
                "Segment does not start with TDSm, but with %s" % s)

        log.debug("Reading segment at %d" % self.position)

        # Next four bytes are table of contents mask
        s = f.read(4)
        toc_mask = struct.unpack('<i', s)[0]

        self.toc = OrderedDict()
        for property in tocProperties.keys():
            self.toc[property] = (toc_mask & tocProperties[property]) != 0
            log.debug("Property %s is %s" % (property, self.toc[property]))

        # Next four bytes are version number
        s = f.read(4)
        self.version = struct.unpack('<i', s)[0]
        if self.version not in (4712, 4713):
            log.warning("Unrecognised version number.")

        # Now 8 bytes each for the offset values
        s = f.read(16)
        (self.next_segment_offset, self.raw_data_offset) = (
            struct.unpack('<QQ', s))

        # Calculate data and next segment position
        lead_size = 7 * 4
        self.data_position = self.position + lead_size + self.raw_data_offset
        if self.next_segment_offset == struct.unpack('<Q', b'\xFF' * 8)[0]:
            # This can happen if Labview crashes
            log.warning(
                "Last segment of file has unknown size, "
                "not attempting to read it")
            self.next_segment_pos = None
            self.next_segment_offset = None
            # Could try to read as much as possible but for now
            # don't attempt to read last segment
            raise EOFError
        else:
            self.next_segment_pos = (
                self.position + self.next_segment_offset + lead_size)

    def __repr__(self):
        return "<TdmsSegment at position %d>" % self.position

    def read_metadata(self, f, objects, previous_segment=None):
        """Read segment metadata section and update object information"""

        if not self.toc["kTocMetaData"]:
            try:
                self.ordered_objects = previous_segment.ordered_objects
            except AttributeError:
                raise ValueError(
                    "kTocMetaData is not set for segment but "
                    "there is no previous segment")
            self.calculate_chunks()
            return
        if not self.toc["kTocNewObjList"]:
            # In this case, there can be a list of new objects that
            # are appended, or previous objects can also be repeated
            # if their properties change
            self.ordered_objects = [
                copy(o) for o in previous_segment.ordered_objects]

        log.debug("Reading metadata at %d" % f.tell())

        # First four bytes have number of objects in metadata
        s = f.read(4)
        num_objects = struct.unpack("<l", s)[0]

        for obj in range(num_objects):
            # Read the object path
            object_path = read_string(f)

            # If this is a new segment for an existing object,
            # reuse the existing object, otherwise,
            # create a new object and add it to the object dictionary
            if object_path in objects:
                obj = objects[object_path]
            else:
                obj = TdmsObject(object_path)
                objects[object_path] = obj

            # Add this segment object to the list of segment objects,
            # re-using any properties from previous segments.
            updating_existing = False
            if not self.toc["kTocNewObjList"]:
                # Search for the same object from the previous segment
                # object list.
                obj_index = [
                    i for i, o in enumerate(self.ordered_objects)
                    if o.tdms_object is obj]
                if len(obj_index) > 0:
                    updating_existing = True
                    log.debug("Updating object in segment list")
                    obj_index = obj_index[0]
                    segment_obj = self.ordered_objects[obj_index]
            if not updating_existing:
                if obj._previous_segment_object is not None:
                    log.debug("Copying previous segment object")
                    segment_obj = copy(obj._previous_segment_object)
                else:
                    log.debug("Creating a new segment object")
                    segment_obj = _TdmsSegmentObject(obj)
                self.ordered_objects.append(segment_obj)
            # Read the metadata for this object, updating any
            # data structure information and properties.
            segment_obj._read_metadata(f)
            obj._previous_segment_object = segment_obj

        self.calculate_chunks()

    def calculate_chunks(self):
        """
        Work out the number of chunks the data is in, for cases
        where the meta data doesn't change at all so there is no
        lead in.

        Also increments the number of values for objects in this
        segment, based on the number of chunks.
        """

        data_size = sum([
            o.data_size
            for o in self.ordered_objects if o.has_data])
        total_data_size = self.next_segment_offset - self.raw_data_offset
        if data_size < 0 or total_data_size < 0:
            raise ValueError("Negative data size")
        elif data_size == 0:
            # Sometimes kTocRawData is set, but there isn't actually any data
            if total_data_size != data_size:
                raise ValueError(
                    "Zero channel data size but non-zero data "
                    "length based on segment offset.")
            self.num_chunks = 0
            return
        if total_data_size % data_size != 0:
            raise ValueError(
                "Data size %d is not a multiple of the "
                "chunk size %d" % (total_data_size, data_size))
        else:
            self.num_chunks = total_data_size // data_size

        # Update data count for the overall tdms object
        # using the data count for this segment.
        for obj in self.ordered_objects:
            if obj.has_data:
                obj.tdms_object.number_values += (
                    obj.number_values * self.num_chunks)

    def read_raw_data(self, f):
        """Read signal data from file"""

        if not self.toc["kTocRawData"]:
            return

        f.seek(self.data_position)

        total_data_size = self.next_segment_offset - self.raw_data_offset
        log.debug(
            "Reading %d bytes of data at %d in %d chunks" %
            (total_data_size, f.tell(), self.num_chunks))

        if self.toc['kTocBigEndian']:
            endianness = '>'
        else:
            endianness = '<'

        for chunk in range(self.num_chunks):
            if self.toc["kTocInterleavedData"]:
                log.debug("Data is interleaved")
                data_objects = [o for o in self.ordered_objects if o.has_data]
                # If all data types have numpy types and all the lengths are
                # the same, then we can read all data at once with numpy,
                # which is much faster
                all_numpy = all(
                    (o.data_type.nptype is not None for o in data_objects))
                same_length = (len(
                    set((o.number_values for o in data_objects))) == 1)
                if (all_numpy and same_length):
                    self._read_interleaved_numpy(f, data_objects, endianness)
                else:
                    self._read_interleaved(f, data_objects, endianness)
            else:
                object_data = {}
                log.debug("Data is contiguous")
                for obj in self.ordered_objects:
                    if obj.has_data:
                        object_data[obj.path] = (
                            obj._read_values(f, endianness))

                for obj in self.ordered_objects:
                    if obj.has_data:
                        obj.tdms_object._update_data(object_data[obj.path])

    def _read_interleaved_numpy(self, f, data_objects, endianness):
        """Read interleaved data where all channels have a numpy type"""

        log.debug("Reading interleaved data all at once")
        # Read all data into 1 byte unsigned ints first
        all_channel_bytes = sum((o.data_type.length for o in data_objects))
        number_bytes = all_channel_bytes * data_objects[0].number_values
        combined_data = np.fromfile(f, dtype=np.uint8, count=number_bytes)
        # Reshape, so that one row is all bytes for all objects
        combined_data = combined_data.reshape(-1, all_channel_bytes)
        # Now set arrays for each channel
        data_pos = 0
        for (i, obj) in enumerate(data_objects):
            byte_columns = tuple(
                range(data_pos, obj.data_type.length + data_pos))
            log.debug("Byte columns for channel %d: %s" % (i, byte_columns))
            # Select columns for this channel, so that number of values will
            # be number of bytes per point * number of data points
            object_data = combined_data[:, byte_columns].ravel()
            # Now set correct data type, so that the array length should
            # be correct
            object_data.dtype = (
                np.dtype(obj.data_type.nptype).newbyteorder(endianness))
            obj.tdms_object._update_data(object_data)
            data_pos += obj.data_type.length

    def _read_interleaved(self, f, data_objects, endianness):
        """Read interleaved data that doesn't have a numpy type"""

        log.debug("Reading interleaved data point by point")
        object_data = {}
        points_added = {}
        for obj in data_objects:
            object_data[obj.path] = obj._new_segment_data()
            points_added[obj.path] = 0
        while any([points_added[o.path] < o.number_values
                  for o in data_objects]):
            for obj in data_objects:
                if points_added[obj.path] < obj.number_values:
                    object_data[obj.path][points_added[obj.path]] = (
                        obj._read_value(f, endianness))
                    points_added[obj.path] += 1
        for obj in data_objects:
            obj.tdms_object._update_data(object_data[obj.path])


class TdmsObject(object):
    """Represents an object in a TDMS file.

    :ivar path: The TDMS object path.
    :ivar properties: Dictionary of TDMS properties defined for this object,
                      for example the start time and time increment for
                      waveforms.
    :ivar has_data: Boolean, true if there is data associated with the object.
    :ivar data: NumPy array containing data if there is data, otherwise None.

    """

    def __init__(self, path):
        self.path = path
        self.data = None
        self.properties = OrderedDict()
        self.data_type = None
        self.dimension = 1
        self.number_values = 0
        self.has_data = False
        self._previous_segment_object = None

    def __repr__(self):
        return "<TdmsObject with path %s>" % self.path

    def property(self, property_name):
        """Returns the value of a TDMS property

        :param property_name: The name of the property to get.
        :returns: The value of the requested property.
        :raises: KeyError if the property isn't found.

        """

        try:
            return self.properties[property_name]
        except KeyError:
            raise KeyError(
                "Object does not have property '%s'" % property_name)

    def time_track(self, absolute_time=False, accuracy='ns'):
        """Return an array of time for this channel

        This depends on the object having the wf_increment
        and wf_start_offset properties defined.

        For larger timespans, the accuracy setting should be set lower.
        The default setting is 'ns', which has a timespan of
        [1678 AD, 2262 AD]. For the exact ranges, refer to
            http://docs.scipy.org/doc/numpy/reference/arrays.datetime.html
        section "Datetime Units".

        :param absolute_time: Whether the returned time values are absolute
            times rather than relative to the start time.
        :param accuracy: The accuracy of the returned datetime64 array.
        :rtype: NumPy array.
        :raises: KeyError if required properties aren't found

        """

        try:
            increment = self.property('wf_increment')
            offset = self.property('wf_start_offset')
        except KeyError:
            raise KeyError("Object does not have time properties available.")

        periods = len(self.data)

        relative_time = np.linspace(
            offset,
            offset + (periods - 1) * increment,
            periods)

        if not absolute_time:
            return relative_time

        try:
            start_time = self.property('wf_start_time')
        except KeyError:
            raise KeyError(
                "Object does not have start time property available.")

        try:
            unit_correction = {
                's': 1e0,
                'ms': 1e3,
                'us': 1e6,
                'ns': 1e9,
            }[accuracy]
        except KeyError:
            raise KeyError("Invalid accuracy: {0}".format(accuracy))

        # Because numpy only knows ints as its date datatype,
        # convert to accuracy.
        time_type = "timedelta64[{0}]".format(accuracy)
        return (np.datetime64(start_time)
                + (relative_time * unit_correction).astype(time_type))

    def _initialise_data(self, memmap_dir=None):
        """Initialise data array to zeros"""

        if self.number_values == 0:
            pass
        elif self.data_type.nptype is None:
            self.data = []
        else:
            if memmap_dir:
                memmap_file = tempfile.NamedTemporaryFile(
                    mode='w+b', prefix="nptdms_", dir=memmap_dir)
                self.data = np.memmap(
                    memmap_file.file,
                    mode='w+',
                    shape=(self.number_values,),
                    dtype=self.data_type.nptype)
            else:
                self.data = np.zeros(
                    self.number_values, dtype=self.data_type.nptype)
            self._data_insert_position = 0

    def _update_data(self, new_data):
        """Update the object data with a new array of data"""

        log.debug("Adding %d data points to data for %s" %
                  (len(new_data), self.path))
        if self.data is None:
            self.data = new_data
        else:
            if self.data_type.nptype is not None:
                data_pos = (
                    self._data_insert_position,
                    self._data_insert_position + len(new_data))
                self._data_insert_position += len(new_data)
                self.data[data_pos[0]:data_pos[1]] = new_data
            else:
                self.data.extend(new_data)

    def as_dataframe(self, absolute_time=False):
        """
        Converts the TDMS object to a DataFrame

        :param absolute_time: Whether times should be absolute rather than
            relative to the start time.
        :return: The TDMS object data.
        :rtype: pandas.DataFrame
        """

        import pandas as pd

        # When absolute_time is True,
        # use the wf_start_time as offset for the time_track()
        time = self.time_track(absolute_time)

        return pd.DataFrame(self.data, index=time, columns=[self.path])


class _TdmsSegmentObject(object):
    """
    Describes an object in an individual TDMS file segment
    """

    __slots__ = [
        'tdms_object', 'number_values', 'data_size',
        'has_data', 'data_type', 'dimension']

    def __init__(self, tdms_object):
        self.tdms_object = tdms_object

        self.number_values = 0
        self.data_size = 0
        self.has_data = True
        self.data_type = None
        self.dimension = 1

    def _read_metadata(self, f):
        """Read object metadata and update object information"""

        s = f.read(4)
        raw_data_index = struct.unpack("<L", s)[0]

        log.debug("Reading metadata for object %s" % self.tdms_object.path)

        # Object has no data in this segment
        if raw_data_index == 0xFFFFFFFF:
            log.debug("Object has no data in this segment")
            self.has_data = False
            # Leave number_values and data_size as set previously,
            # as these may be re-used by later segments.
        # Data has same structure as previously
        elif raw_data_index == 0x00000000:
            log.debug(
                "Object has same data structure as in the previous segment")
            self.has_data = True
        else:
            # raw_data_index gives the length of the index information.
            self.has_data = True
            self.tdms_object.has_data = True

            # Read the data type
            s = f.read(4)
            try:
                self.data_type = tdsDataTypes[struct.unpack("<L", s)[0]]
            except KeyError:
                raise KeyError("Unrecognised data type")
            if (self.tdms_object.data_type is not None and
                    self.data_type != self.tdms_object.data_type):
                raise ValueError(
                    "Segment object doesn't have the same data "
                    "type as previous segments.")
            else:
                self.tdms_object.data_type = self.data_type
            log.debug("Object data type: %s" % self.tdms_object.data_type.name)

            if (self.tdms_object.data_type.length is None and
                    self.tdms_object.data_type.name != 'tdsTypeString'):
                raise ValueError(
                    "Unsupported data type: %s" %
                    self.tdms_object.data_type.name)

            # Read data dimension
            s = f.read(4)
            self.dimension = struct.unpack("<L", s)[0]
            if self.dimension != 1:
                log.warning("Data dimension is not 1")

            # Read number of values
            s = f.read(8)
            self.number_values = struct.unpack("<Q", s)[0]

            # Variable length data types have total size
            if self.data_type.name in ('tdsTypeString', ):
                s = f.read(8)
                self.data_size = struct.unpack("<Q", s)[0]
            else:
                self.data_size = (
                    self.number_values *
                    self.data_type.length * self.dimension)

            log.debug(
                "Object number of values in segment: %d" % self.number_values)

        # Read data properties
        s = f.read(4)
        num_properties = struct.unpack("<L", s)[0]
        log.debug("Reading %d properties" % num_properties)
        for i in range(num_properties):
            prop_name = read_string(f)

            # Property data type
            s = f.read(4)
            prop_data_type = tdsDataTypes[struct.unpack("<L", s)[0]]
            if prop_data_type.name == 'tdsTypeString':
                value = read_string(f)
            else:
                value = read_type(f, prop_data_type, '<')
            log.debug("Property %s (%s): %s" % (
                prop_name, prop_data_type.name, value))
            self.tdms_object.properties[prop_name] = value

    @property
    def path(self):
        return self.tdms_object.path

    def _read_value(self, file, endianness):
        """Read a single value from the given file"""

        if self.data_type.nptype is not None:
            dtype = (np.dtype(self.data_type.nptype).newbyteorder(endianness))
            return np.fromfile(file, dtype=dtype, count=1)
        return read_type(file, self.data_type, endianness)

    def _read_values(self, file, endianness):
        """Read all values for this object from a contiguous segment"""

        if self.data_type.nptype is not None:
            dtype = (np.dtype(self.data_type.nptype).newbyteorder(endianness))
            return np.fromfile(file, dtype=dtype, count=self.number_values)
        elif self.data_type.name == "tdsTypeString":
            return read_string_data(file, self.number_values)
        data = self._new_segment_data()
        for i in range(self.number_values):
            data[i] = read_type(file, self.data_type, endianness)
        return data

    def _new_segment_data(self):
        """Return a new array to read the data of the current section into"""

        if self.data_type.nptype is not None:
            return np.zeros(self.number_values, dtype=self.data_type.nptype)
        else:
            return [None] * self.number_values


def read_string_data(file, number_values):
    """ Read string raw data

        This is stored as an array of offsets
        followed by the contiguous string data.
    """
    offsets = [0]
    for i in range(number_values):
        s = file.read(4)
        offset = struct.unpack("<L", s)[0]
        offsets.append(offset)
    strings = []
    for i in range(number_values):
        s = file.read(offsets[i + 1] - offsets[i])
        strings.append(s.decode('utf-8'))
    return strings



"""Test exporting TDMS data to Pandas"""

import unittest
import sys
import logging
from datetime import datetime
try:
    import pytz
except ImportError:
    pytz = None

from nptdms import tdms
from .tdms_test import (
    TestFile,
    basic_segment,
    string_hexlify,
    hexlify_value
)


if pytz:
    timezone = pytz.utc
else:
    timezone = None


def within_tol(a, b, tol=1.0e-10):
    return abs(a - b) < tol


def timed_segment():
    """TDMS segment with one group and two channels,
    each with time properties"""

    toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList")
    metadata = (
        # Number of objects
        "03 00 00 00"
        # Length of the first object path
        "08 00 00 00"
        # Object path (/'Group')
        "2F 27 47 72"
        "6F 75 70 27"
        # Raw data index
        "FF FF FF FF"
        # Num properties
        "02 00 00 00"
        # Name length
        "04 00 00 00"
        # Property name (prop)
        "70 72 6F 70"
        # Property data type (string)
        "20 00 00 00"
        # Length of string value
        "05 00 00 00"
        # Value
        "76 61 6C 75 65"
        # Length of second property name
        "03 00 00 00"
        # Property name (num)
        "6E 75 6D"
        # Data type of property
        "03 00 00 00"
        # Value
        "0A 00 00 00"
        # Length of the second object path
        "13 00 00 00"
        # Second object path (/'Group'/'Channel1')
        "2F 27 47 72"
        "6F 75 70 27"
        "2F 27 43 68"
        "61 6E 6E 65"
        "6C 31 27"
        # Length of index information
        "14 00 00 00"
        # Raw data data type
        "03 00 00 00"
        # Dimension
        "01 00 00 00"
        # Number of raw datata values
        "02 00 00 00"
        "00 00 00 00"
        # Number of properties
        "03 00 00 00"
        # Set time properties for the first channel
        "0F 00 00 00" +
        string_hexlify('wf_start_offset') +
        "0A 00 00 00" +
        hexlify_value("<d", 2.0) +
        "0C 00 00 00" +
        string_hexlify('wf_increment') +
        "0A 00 00 00" +
        hexlify_value("<d", 0.1) +
        "0D 00 00 00" +
        string_hexlify('wf_start_time') +
        "44 00 00 00" +
        hexlify_value("<Q", 0) +
        hexlify_value("<q", 3524551547) +
        # Length of the third object path
        "13 00 00 00"
        # Third object path (/'Group'/'Channel2')
        "2F 27 47 72"
        "6F 75 70 27"
        "2F 27 43 68"
        "61 6E 6E 65"
        "6C 32 27"
        # Length of index information
        "14 00 00 00"
        # Raw data data type
        "03 00 00 00"
        # Dimension
        "01 00 00 00"
        # Number of data values
        "02 00 00 00"
        "00 00 00 00"
        # Number of properties
        "03 00 00 00"
        # Set time properties for the second channel
        "0F 00 00 00" +
        string_hexlify('wf_start_offset') +
        "0A 00 00 00" +
        hexlify_value("<d", 2.0) +
        "0C 00 00 00" +
        string_hexlify('wf_increment') +
        "0A 00 00 00" +
        hexlify_value("<d", 0.1) +
        "0D 00 00 00" +
        string_hexlify('wf_start_time') +
        "44 00 00 00" +
        hexlify_value("<Q", 0) +
        hexlify_value("<q", 3524551547))
    data = (
        # Data for segment
        "01 00 00 00"
        "02 00 00 00"
        "03 00 00 00"
        "04 00 00 00"
    )
    return (metadata, data, toc)


class TDMSTestClass(unittest.TestCase):
    def setUp(self):
        logging.getLogger(tdms.__name__).setLevel(logging.DEBUG)

    def test_file_as_dataframe(self):
        """Test converting file to Pandas dataframe"""

        test_file = TestFile()
        test_file.add_segment(*timed_segment())
        tdmsData = test_file.load()

        df = tdmsData.as_dataframe()

        self.assertEqual(len(df), 2)
        self.assertIn("/'Group'/'Channel1'", df.keys())
        self.assertIn("/'Group'/'Channel2'", df.keys())

        self.assertTrue((df["/'Group'/'Channel1'"] == [1, 2]).all())

    def test_file_as_dataframe_without_time(self):
        """Converting file to dataframe with time index should raise when
        time properties aren't present"""

        test_file = TestFile()
        test_file.add_segment(*basic_segment())
        tdmsData = test_file.load()

        with self.assertRaises(KeyError):
            df = tdmsData.as_dataframe(time_index=True)

    def test_file_as_dataframe_with_time(self):
        """Test converting file to Pandas dataframe with a time index"""

        test_file = TestFile()
        test_file.add_segment(*timed_segment())
        tdmsData = test_file.load()

        df = tdmsData.as_dataframe(time_index=True)

        self.assertEqual(len(df.index), 2)
        self.assertTrue(within_tol(df.index[0], 2.0))
        self.assertTrue(within_tol(df.index[1], 2.1))

    def test_file_as_dataframe_with_absolute_time(self):
        """Convert file to Pandas dataframe with absolute time index"""

        test_file = TestFile()
        test_file.add_segment(*timed_segment())
        tdmsData = test_file.load()

        df = tdmsData.as_dataframe(time_index=True, absolute_time=True)

        expected_start = datetime(2015, 9, 8, 10, 5, 49, tzinfo=timezone)
        self.assertTrue((df.index == expected_start)[0])

    def test_channel_as_dataframe(self):
        """Convert a channel to dataframe"""

        test_file = TestFile()
        test_file.add_segment(*timed_segment())
        tdmsData = test_file.load()

        df = tdmsData.object("Group", "Channel2").as_dataframe()
        self.assertEqual(len(df), 2)
        self.assertEqual(len(df.keys()), 1)
        self.assertIn("/'Group'/'Channel2'", df.keys())
        self.assertTrue((df["/'Group'/'Channel2'"] == [3, 4]).all())

    def test_channel_as_dataframe_with_time(self):
        """Convert a channel to dataframe with a time index"""

        test_file = TestFile()
        test_file.add_segment(*timed_segment())
        tdmsData = test_file.load()

        df = tdmsData.object("Group", "Channel2").as_dataframe()
        self.assertEqual(len(df.index), 2)
        self.assertTrue(within_tol(df.index[0], 2.0))
        self.assertTrue(within_tol(df.index[1], 2.1))

    def test_channel_as_dataframe_without_time(self):
        """Converting channel to dataframe should raise when
        time properties aren't present"""

        test_file = TestFile()
        test_file.add_segment(*basic_segment())
        tdmsData = test_file.load()

        with self.assertRaises(KeyError):
            df = tdmsData.object("Group", "Channel1").as_dataframe()

    def test_channel_as_dataframe_with_absolute_time(self):
        """Convert channel to Pandas dataframe with absolute time index"""

        test_file = TestFile()
        test_file.add_segment(*timed_segment())
        tdmsData = test_file.load()

        df = tdmsData.object("Group", "Channel1").as_dataframe(
            absolute_time=True)

        expected_start = datetime(2015, 9, 8, 10, 5, 49, tzinfo=timezone)
        self.assertTrue((df.index == expected_start)[0])


if __name__ == '__main__':
    unittest.main()
